import os
from mpcontribs.io.archieml.mpfile import MPFile

def test_get_string():
    test_file = os.path.join(os.path.dirname(__file__), 'test_archieml.txt')
    mpfile = MPFile.from_file(test_file)
    mpfile_test = MPFile.from_string(mpfile.get_string())
    assert mpfile.document == mpfile_test.document



# -*- coding: utf-8 -*-
"""module defines utility methods for MPContribs core.io library"""
from __future__ import unicode_literals
import collections
from decimal import Decimal, DecimalException, InvalidOperation
import six
from mpcontribs.io import mp_id_pattern

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

def get_short_object_id(cid):
    """return shortened contribution ID (ObjectId) for `cid`.

    >>> get_short_object_id('5a8638add4f144413451852a')
    '451852a'
    >>> get_short_object_id('5a8638add4f1400000000000')
    '5a8638a'
    """
    length = 7
    cid_short = str(cid)[-length:]
    if cid_short == '0'*length:
        cid_short = str(cid)[:length]
    return cid_short

def make_pair(key, value, sep=':'):
    """return string for `key`-`value` pair with separator `sep`.

    >>> make_pair('Phase', 'Hollandite')
    u'Phase: Hollandite'
    >>> print make_pair('ΔH', '0.066 eV/mol', sep=';')
    ΔH; 0.066 eV/mol
    >>> make_pair('k', 2.3)
    u'k: 2.3'
    """
    if not isinstance(value, six.string_types):
        value = unicode(value)
    return '{} '.format(sep).join([key, value])

def nest_dict(dct, keys):
    #"""nest `dct` under list of `keys`.

    #>>> print nest_dict({'key': {'subkey': 'value'}}, ['a', 'b'])
    #RecursiveDict([('a', RecursiveDict([('b', RecursiveDict([('key', RecursiveDict([('subkey', u'value')]))]))]))])
    #"""
    from mpcontribs.io.core.recdict import RecursiveDict
    nested_dict = dct
    #nested_dict = RecursiveDict(dct)
    #nested_dict.rec_update()
    for key in reversed(keys):
        nested_dict = RecursiveDict({key: nested_dict})
    return nested_dict

def get_composition_from_string(comp_str):
    """validate and return composition from string `comp_str`."""
    from pymatgen import Composition, Element
    comp = Composition(comp_str)
    for element in comp.elements:
        Element(element)
    formula = comp.get_integer_formula_and_factor()[0]
    comp = Composition(formula)
    return ''.join([
        '{}{}'.format(key, int(value) if value > 1 else '')
        for key, value in comp.as_dict().items()
    ])

def normalize_root_level(title):
    """convert root-level title into conventional identifier; non-identifiers
    become part of shared (meta-)data. Returns: (is_general, title)"""
    from pymatgen.core.composition import CompositionError
    try:
        composition = get_composition_from_string(title)
        return False, composition
    except (CompositionError, KeyError, TypeError, ValueError):
        if mp_id_pattern.match(title.lower()):
            return False, title.lower()
        return True, title

def clean_value(value, unit='', convert_to_percent=False, max_dgts=3):
    """return clean value with maximum digits and optional unit and percent"""
    dgts = max_dgts
    value = str(value) if not isinstance(value, six.string_types) else value
    try:
        value = Decimal(value)
        dgts = len(value.as_tuple().digits)
        dgts = max_dgts if dgts > max_dgts else dgts
    except DecimalException:
        return value
    if convert_to_percent:
        value = Decimal(value) * Decimal('100')
        unit = '%'
    val = '{{:.{}g}}'.format(dgts).format(value)
    if unit:
        val += ' {}'.format(unit)
    return val

def strip_converter(text):
    """http://stackoverflow.com/questions/13385860"""
    try:
        text = text.strip()
        if not text:
            return ''
        val = clean_value(text, max_dgts=6)
        return str(Decimal(val))
    except InvalidOperation:
        return text

def read_csv(body, is_data_section=True, **kwargs):
    """run pandas.read_csv on (sub)section body"""
    csv_comment_char = '#'
    import pandas
    body = body.strip()
    if not body:
        return None
    from mpcontribs.io.core.components.tdata import Table
    if is_data_section:
        cur_line = 1
        while 1:
            body_split = body.split('\n', cur_line)
            first_line = body_split[cur_line-1].strip()
            cur_line += 1
            if first_line and not first_line.startswith(csv_comment_char):
                break
        sep = kwargs.get('sep', ',')
        options = {'sep': sep, 'header': 0}
        header = [col.strip() for col in first_line.split(sep)]
        body = '\n'.join([sep.join(header), body_split[1]])
        if first_line.startswith('level_'):
            options.update({'index_col': [0, 1]})
        ncols = len(header)
    else:
        options = {'sep': ':', 'header': None, 'index_col': 0}
        ncols = 2
    options.update(**kwargs)
    converters = dict((col, strip_converter) for col in range(ncols))
    return Table(pandas.read_csv(
        StringIO(body), comment=csv_comment_char,
        skipinitialspace=True, squeeze=True,
        converters=converters, encoding='utf8',
        **options
    ).dropna(how='all'))



from IPython.display import display_html
from mpcontribs.io import mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from mpcontribs.io.core.utils import nest_dict

class HierarchicalData(RecursiveDict):
    """class to hold and display all hierarchical data in MPFile"""
    def __init__(self, doc):
        super(HierarchicalData, self).__init__()
        document = RecursiveDict(doc)
        scope = []
        for key, value in document.iterate():
            level, key = key
            if key in mp_level01_titles:
                continue
            level_reduction = bool(level < len(scope))
            if level_reduction:
                del scope[level:]
            if value is None:
                scope.append(key)
            else:
                d = nest_dict(value, scope + [key])
                self.rec_update(d, overwrite=True)



from __future__ import unicode_literals, print_function
import archieml, textwrap
from mpcontribs.io import replacements, mp_level01_titles
from mpcontribs.io.core.mpfile import MPFileCore
from mpcontribs.io.core.recdict import RecursiveDict, Quantity
from mpcontribs.io.core.utils import nest_dict, normalize_root_level
from mpcontribs.io.core.utils import read_csv, make_pair
from mpcontribs.io.core.components.tdata import Table
from pandas import MultiIndex

class MPFile(MPFileCore):
    """Object for representing a MP Contribution File in ArchieML format."""

    @staticmethod
    def from_string(data):
        # use archieml-python parse to import data
        rdct = RecursiveDict(archieml.loads(data))
        rdct.rec_update()

        # post-process internal representation of file contents
        for key in list(rdct.keys()):
            is_general, root_key = normalize_root_level(key)

            if is_general:
                # make part of shared (meta-)data, i.e. nest under `general` at
                # the beginning of the MPFile
                if mp_level01_titles[0] not in rdct:
                    rdct[mp_level01_titles[0]] = RecursiveDict()
                    rdct.move_to_end(mp_level01_titles[0], last=False)

            # normalize identifier key (pop & insert)
            # using rec_update since we're looping over all entries
            # also: support data in bare tables (marked-up only by
            #       root-level identifier) by nesting under 'data'
            value = rdct.pop(key)
            keys = [mp_level01_titles[0]] if is_general else []
            keys.append(root_key)
            if isinstance(value, list):
                keys.append('table')
            rdct.rec_update(nest_dict(value, keys))

            # reference to section to iterate or parse as CIF
            section = rdct[mp_level01_titles[0]][root_key] \
                    if is_general else rdct[root_key]

            # iterate to find CSV sections to parse
            # also parse propnet quantities
            if isinstance(section, dict):
                scope = []
                for k, v in section.iterate():
                    level, key = k
                    key = ''.join([replacements.get(c, c) for c in key])
                    level_reduction = bool(level < len(scope))
                    if level_reduction:
                        del scope[level:]
                    if v is None:
                        scope.append(key)
                    elif isinstance(v, list) and isinstance(v[0], dict):
                        table = ''
                        for row_dct in v:
                            table = '\n'.join([table, row_dct['value']])
                        pd_obj = read_csv(table)
                        d = nest_dict(pd_obj.to_dict(), scope + [key])
                        section.rec_update(d, overwrite=True)
                        if not is_general and level == 0:
                            section.insert_default_plot_options(pd_obj, key)
                    elif Quantity is not None and isinstance(v, six.string_types) and ' ' in v:
                        quantity = Quantity.from_key_value(key, v)
                        d = nest_dict(quantity.as_dict(), scope + [key]) # TODO quantity.symbol.name
                        section.rec_update(d, overwrite=True)


            # convert CIF strings into pymatgen structures
            if mp_level01_titles[3] in section:
                from pymatgen.io.cif import CifParser
                for name in section[mp_level01_titles[3]].keys():
                    cif = section[mp_level01_titles[3]].pop(name)
                    parser = CifParser.from_string(cif)
                    structure = parser.get_structures(primitive=False)[0]
                    section[mp_level01_titles[3]].rec_update(nest_dict(
                        structure.as_dict(), [name]
                    ))

        return MPFile.from_dict(rdct)

    def get_string(self, df_head_only=False):
        from pymatgen import Structure
        lines, scope = [], []
        for key,value in self.document.iterate():
            if isinstance(value, Table):
                lines[-1] = lines[-1].replace('{', '[+').replace('}', ']')
                header = any([isinstance(col, str) for col in value])
                if isinstance(value.index, MultiIndex):
                    value.reset_index(inplace=True)
                if df_head_only:
                    value = value.head()
                csv_string = value.to_csv(
                    index=False, header=header, float_format='%g', encoding='utf-8'
                )[:-1]
                lines += csv_string.split('\n')
                if df_head_only:
                    lines.append('...')
            elif isinstance(value, Structure):
                from pymatgen.io.cif import CifWriter
                cif = CifWriter(value, symprec=1e-10).__str__()
                lines.append(make_pair(
                    ''.join([replacements.get(c, c) for c in key]), cif+':end'
                ))
            elif Quantity is not None and isinstance(value, Quantity):
                lines.append(make_pair(
                    value.display_symbols[0], value.pretty_string()
                ))
            else:
                level, key = key
                # truncate scope
                level_reduction = bool(level < len(scope))
                if level_reduction:
                    del scope[level:]
                # append scope
                if value is None:
                    scope.append(''.join([
                        replacements.get(c, c) for c in key
                    ]))
                # correct scope to omit internal 'general' section
                scope_corr = scope
                if scope[0] == mp_level01_titles[0]:
                    scope_corr = scope[1:]
                # insert scope line
                if (value is None and scope_corr) or \
                   (value is not None and level_reduction):
                    lines.append('\n{' + '.'.join(scope_corr) + '}')
                # insert key-value line
                if value is not None:
                    val = str(value)
                    value_lines = [val] if val.startswith('http') \
                            else textwrap.wrap(val)
                    if len(value_lines) > 1:
                        value_lines = [''] + value_lines + [':end']
                    lines.append(make_pair(
                        ''.join([replacements.get(c, c) for c in key]),
                        '\n'.join(value_lines)
                    ))
        return '\n'.join(lines) + '\n'

MPFileCore.register(MPFile)



from __future__ import unicode_literals, print_function
import six, codecs, os
from abc import ABCMeta
from tempfile import gettempdir
from mpcontribs.io import replacements, mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from mpcontribs.io.core.utils import nest_dict, get_composition_from_string
from mpcontribs.io.core.components.hdata import HierarchicalData
from mpcontribs.io.core.components.tdata import TabularData, Table
from mpcontribs.io.core.components.gdata import GraphicalData
from mpcontribs.io.core.components.sdata import StructuralData

default_mpfile_path = os.path.join(gettempdir(), 'mpfile.txt')

class MPFileCore(six.with_metaclass(ABCMeta, object)):
    """Abstract Base Class for representing a MP Contribution File"""
    def __init__(self, data=RecursiveDict()):
        if isinstance(data, dict):
            self.document = RecursiveDict(data)
        else:
            raise ValueError('Need dict (or inherited class) to init MPFile.')
        self.document.rec_update() # convert (most) OrderedDict's to RecursiveDict's
        self.unique_mp_cat_ids = True
        self.max_contribs = 10

    def __getitem__(self, key):
        item = self.from_dict({key: self.document[key]})
        general = self.document.get(mp_level01_titles[0])
        if general:
            item.insert_general_section(self.from_dict({mp_level01_titles[0]: general}))
        return item

    @property
    def ids(self):
        return [
            k for k in self.document.keys()
            if k.lower() != mp_level01_titles[0]
        ]

    @property
    def hdata(self):
        return HierarchicalData(self.document)

    @property
    def tdata(self):
        return TabularData(self.document)

    @property
    def gdata(self):
        return GraphicalData(self.document)

    @property
    def sdata(self):
        return StructuralData(self.document)

    @classmethod
    def from_file(cls, filename_or_file=default_mpfile_path.replace('.txt', '_in.txt')):
        """Reads a MPFile from a file.

        Args:
            filename_or_file (str or file): name of file or file containing contribution data.

        Returns:
            MPFile object.
        """
        f = open(filename_or_file) \
            if isinstance(filename_or_file, six.string_types) \
            else filename_or_file
        return cls.from_string(f.read())

    @classmethod
    def from_dict(cls, data=RecursiveDict()):
        return cls(data=data)

    @classmethod
    def from_contribution(cls, contrib):
        """construct MPFile from contribution (see rest.adapter.submit_contribution)"""
        if not 'identifier' in contrib or not 'content' in contrib:
            raise ValueError('Dict not in contribution-style format')
        recdict = RecursiveDict({contrib['identifier']: contrib['content']})
        return cls.from_dict(recdict)

    def write_file(self, filename=default_mpfile_path.replace('.txt', '_out.txt'), **kwargs):
        """Writes MPFile to a file. The supported kwargs are the same as those
        for the MPFile.get_string method and are passed through directly."""
        with codecs.open(filename, encoding='utf-8', mode='w') as f:
            file_str = self.get_string(**kwargs) + '\n'
            f.write(file_str)
            print('{} ({:.3f}MB) written'.format(
                filename, os.path.getsize(filename) / 1024. / 1024.
            ))

    def get_number_of_lines(self, **kwargs):
        return len(self.get_string(**kwargs).split('\n'))

    def split(self):
        general_mpfile = self.pop_first_section() \
                if mp_level01_titles[0] in self.document.keys() else None
        if not self.document:
            raise ValueError('No contributions in MPFile! Either the file is'
                             ' empty or only contains shared (meta-)data not'
                             ' correlated to core identifier.')
        while True:
            try:
                mpfile_single = self.pop_first_section()
                mpid_orig = mpfile_single.ids[0]
                if '--' in mpid_orig:
                    mpid = mpid_orig.split('--')[0]
                    mpfile_single.document.rec_update(nest_dict(
                        mpfile_single.document.pop(mpid_orig), [mpid]
                    ))
                if general_mpfile is not None:
                    mpfile_single.insert_general_section(general_mpfile)
                yield mpfile_single
            except KeyError:
                break

    def get_identifiers(self):
        """list of materials/composition identifiers as tuples w/ contribution IDs"""
        return [
            (k, self.document[k].get('cid', None))
            for k in self.document
            if k.lower() != mp_level01_titles[0]
        ]

    def pop_first_section(self):
        item = self.document.popitem(last=False)
        return self.from_dict(RecursiveDict([item]))

    def insert_general_section(self, general_mpfile):
        """insert general section from `general_mpfile` into this MPFile"""
        if general_mpfile is None:
            return
        general_title = mp_level01_titles[0]
        general_data = general_mpfile.document[general_title]
        root_key = list(self.document.keys())[0]
        for key, value in general_data.items():
            if key in self.document[root_key]:
                self.document.rec_update(nest_dict(value, [root_key, key]))
            else:
                self.document[root_key][key] = value
        for key in reversed(general_data.keys()):
            self.document[root_key].move_to_end(key, last=False)

    def get_unique_mp_cat_id(self, mp_cat_id):
        if not self.unique_mp_cat_ids or mp_cat_id in mp_level01_titles:
            return mp_cat_id
        mp_cat_id_idx = len([i for i in self.ids if i.startswith(mp_cat_id)])
        if mp_cat_id_idx == 0:
            return mp_cat_id
        return '{}--{}'.format(mp_cat_id, mp_cat_id_idx)

    def concat(self, mpfile):
        """concatenate single-section MPFile with this MPFile"""
        try:
            if len(mpfile.document) > 1:
                raise ValueError('concatenation only possible with single section files')
        except AttributeError:
            raise ValueError('Provide a MPFile to concatenate')
        mp_cat_id = list(mpfile.document.keys())[0]
        general_title = mp_level01_titles[0]
        if general_title in mpfile.document[mp_cat_id]:
            general_data = mpfile.document[mp_cat_id].pop(general_title)
            if general_title not in self.document:
                self.document.rec_update(nest_dict(general_data, [general_title]))
        self.document.rec_update(nest_dict(
            mpfile.document.pop(mp_cat_id),
            [self.get_unique_mp_cat_id(mp_cat_id)]
        ))

    def insert_top(self, mp_cat_id, key, value):
        """insert value for `mp_cat_id` as `key: <value>` at top"""
        self.document[mp_cat_id][key] = str(value)
        self.document[mp_cat_id].move_to_end(key, last=False)

    def add_data_table(self, identifier, dataframe, name, plot_options=None):
        """add a datatable to the root-level section

        Args:
            identifier (str): MP category ID (`mp_cat_id`)
            dataframe (pandas.DataFrame): tabular data as Pandas DataFrame
            name (str): table name, optional if only one table in section
            plot_options (dict): options for according plotly graph
        """
        # TODO: optional table name, required if multiple tables per root-level section
        name = ''.join([replacements.get(c, c) for c in name])
        self.document.rec_update(nest_dict(
            Table(dataframe).to_dict(), [identifier, name]
        ))
        self.document[identifier].insert_default_plot_options(
            dataframe, name, update_plot_options=plot_options
        )

    def add_hierarchical_data(self, dct, identifier=mp_level01_titles[0]):
        if len(self.ids) >= self.max_contribs:
            raise StopIteration('Reached max. number of contributions in MPFile')
        self.document.rec_update(nest_dict(RecursiveDict(dct), [identifier]))

    def add_structure(self, source, name=None, identifier=None, fmt=None):
        """add a structure to the mpfile"""
        from pymatgen import Structure, MPRester
        if isinstance(source, Structure):
            structure = source
        elif isinstance(source, dict):
            structure = Structure.from_dict(source)
        elif os.path.exists(source):
            structure = Structure.from_file(source, sort=True)
        elif isinstance(source, six.string_types):
            if fmt is None:
                raise ValueError('Need fmt to get structure from string!')
            structure = Structure.from_str(source, fmt, sort=True)
        else:
            raise ValueError(source, 'not supported!')

        if name is not None:
            if not isinstance(name, six.string_types):
                raise ValueError('structure name needs to be a string')
            elif '.' in name:
                raise ValueError('structure name cannot contain dots (.)')

        mpr = MPRester()
        if not mpr.api_key:
            raise ValueError(
                'API key not set. Run `pmg config --add PMG_MAPI_KEY <USER_API_KEY>`.'
            )
        matched_mpids = mpr.find_structure(structure)
        formula = get_composition_from_string(structure.composition.formula)
        if not matched_mpids:
            if identifier is None:
                identifier = formula
                print(
                    'Structure not found in MP! Please submit via MPComplete to '
                    'obtain mp-id or manually choose an anchor mp-id! Continuing '
                    'with {} as identifier!'.format(identifier)
                )
            else:
                print('Structure not found in MP! Forcing {} as identifier!'.format(identifier))
        elif identifier is None:
            identifier = matched_mpids[0]
            if len(matched_mpids) > 1:
                print('Multiple matching structures found in MP. Using', identifier)
        elif identifier not in matched_mpids:
            msg = 'Structure does not match {} but instead {}!'.format(identifier, matched_mpids)
            raise ValueError(msg)

        idx = len(self.document.get(identifier, {}).get(mp_level01_titles[3], {}))
        sub_key = formula if name is None else name
        if sub_key in self.document.get(identifier, {}).get(mp_level01_titles[3], {}):
            sub_key += '_{}'.format(idx)
        self.document.rec_update(nest_dict(
            structure.as_dict(), [identifier, mp_level01_titles[3], sub_key]
        ))
        return identifier

    def __repr__(self): return self.get_string(df_head_only=True)
    def __str__(self):
        return self.get_string(df_head_only=True)

    def _ipython_display_(self):
        from IPython.display import display_html
        display_html(self.hdata)
        display_html(self.tdata)
        display_html(self.gdata)
        display_html(self.sdata)

    # ----------------------------
    # Override these in subclasses
    # ----------------------------

    @staticmethod
    def from_string(data):
        """Reads a MPFile from a string containing contribution data."""
        return MPFileCore()

    def get_string(self, df_head_only=False):
        """Returns a string to be written as a file"""
        return repr(self.document)



import re
mp_id_pattern = re.compile('^(mp|por|mvc)-\d+(?:--\d+)?$', re.IGNORECASE)
replacements = {' ': '_', '[': '', ']': '', '{': '', '}': '', ':': '_'}
mp_level01_titles = [ '_hdata', 'tables', 'graphs', 'structures' ]



import uuid, json
import pandas as pd
from IPython.display import display_html, HTML
from mpcontribs.io import mp_level01_titles, mp_id_pattern
from mpcontribs.io.core.utils import nest_dict, clean_value
from mpcontribs.io.core.recdict import RecursiveDict
from urllib.parse import urlparse

class Table(pd.DataFrame):
    def __init__(self, data, columns=None, index=None,
                 cid=None, name=None, api_key=None, project=None):
        super(Table, self).__init__(data=data, index=index, columns=columns)
        self.cid = cid
        self.name = name
        self.api_key = api_key
        self.project = project

    def to_dict(self):
        from pandas import MultiIndex
        for col in self.columns:
            self[col] = self[col].apply(lambda x: clean_value(x, max_dgts=6))
        rdct = super(Table, self).to_dict(orient='split', into=RecursiveDict)
        if not isinstance(self.index, MultiIndex):
            rdct.pop('index')
        rdct["@module"] = self.__class__.__module__
        rdct["@class"] = self.__class__.__name__
        return rdct

    @classmethod
    def from_dict(cls, d):
        index = None
        if 'index' in d:
            from pandas import MultiIndex
            index = MultiIndex.from_tuples(d['index'])
        obj = cls(
            d['data'], columns=d['columns'], index=index,
            cid=d['cid'], name=d['name']
        ) if 'cid' in d and 'name' in d else cls(
            d['data'], columns=d['columns'], index=index
        )
        return obj

    @classmethod
    def from_items(cls, rdct, **kwargs):
        return super(Table, cls).from_dict(RecursiveDict(rdct), **kwargs)

    def to_backgrid_dict(self):
        """Backgrid-conform dict from DataFrame"""
        # shorten global import times by importing django here
        import numpy as np
        from mpcontribs.io.core.utils import get_composition_from_string
        from pandas import MultiIndex
        import pymatgen.util as pmg_util
        from pymatgen.core.composition import CompositionError

        table = dict()
        nrows_max = 260
        nrows = self.shape[0]
        df = Table(self.head(n=nrows_max)) if nrows > nrows_max else self

        if isinstance(df.index, MultiIndex):
            df.reset_index(inplace=True)

        table['columns'] = []
        table['rows'] = super(Table, df).to_dict(orient='records')

        for col_index, col in enumerate(list(df.columns)):
            cell_type = 'number'

            # avoid looping rows to minimize use of `df.iat` (time-consuming in 3d)
            if not col.startswith('level_') and col[-1] != ']':
                is_url_column = True

                for row_index in range(df.shape[0]):
                    cell = str(df.iat[row_index, col_index])
                    is_url_column = bool(is_url_column and (not cell or mp_id_pattern.match(cell)))
                    if is_url_column:
                        if cell:
                            value = 'https://materialsproject.org/materials/{}'.format(cell)
                            table['rows'][row_index][col] = value
                    elif cell:
                        try:
                            composition = get_composition_from_string(cell)
                            composition = pmg_util.string.unicodeify(composition)
                            table['rows'][row_index][col] = composition
                        except (CompositionError, ValueError, OverflowError):
                            try:
                                # https://stackoverflow.com/a/38020041
                                result = urlparse(cell)
                                if not all([result.scheme, result.netloc, result.path]):
                                    break
                                is_url_column = True
                            except:
                                break

                cell_type = 'uri' if is_url_column else 'string'

            col_split = col.split('.')
            nesting = [col_split[0]] if len(col_split) > 1 else []
            table['columns'].append({
                'name': col, 'cell': cell_type, 'nesting': nesting, 'editable': 0
            })
            if len(col_split) > 1:
                table['columns'][-1].update({'label': '.'.join(col_split[1:])})

        return table

    def render(self, total_records=None):
        """use BackGrid JS library to render Pandas DataFrame"""
        # if project given, this will result in an overview table of contributions
        # TODO check for index column in df other than the default numbering
        jtable = json.dumps(self.to_backgrid_dict())
        if total_records is None:
            total_records = self.shape[0]
        config = {"total_records": total_records}
        config['uuids'] = [str(uuid.uuid4()) for i in range(4)]
        if self.project is None:
            config['name'] = self.name
            config['cid'] = self.cid
        else:
            config['project'] = self.project
        config['api_key'] = self.api_key
        jconfig = json.dumps(config)
        html = '<div class="col-md-6" id="{}"></div><div class="pull-right" id="{}"></div>'.format(config['uuids'][0], config['uuids'][3])
        html += '<div id="{}" style="width:100%;"></div>'.format(config['uuids'][1])
        html += '<div id="{}"></div>'.format(config['uuids'][2])
        html += f'<script>render_table({{table: {jtable}, config: {jconfig}}})</script>'
        return html

    def _ipython_display_(self):
        display(HTML(self.render()))

class Tables(RecursiveDict):
    """class to hold and display multiple data tables"""
    def __init__(self, content=RecursiveDict()):
        super(Tables, self).__init__(
            (key, value) for key, value in content.items()
            if isinstance(value, Table)
        )

    def __str__(self):
        return 'tables: {}'.format(' '.join(self.keys()))

    def _ipython_display_(self):
        for name, table in self.items():
            display_html('<h3>{}</h3>'.format(name), raw=True)
            display_html(table)

class TabularData(RecursiveDict):
    """class to hold and display all tabular data of a MPFile"""
    def __init__(self, document):
        super(TabularData, self).__init__()
        from pymatgen import Structure
        scope = []
        for key, value in document.iterate():
            if isinstance(value, Table):
                self[scope[0]].rec_update({'.'.join(scope[1:]): value})
            elif not isinstance(value, Structure):
                level, key = key
                level_reduction = bool(level < len(scope))
                if level_reduction:
                    del scope[level:]
                if value is None:
                    scope.append(key)
                    if scope[0] not in self:
                        self[scope[0]] = Tables()

    def __str__(self):
        return 'mp-ids: {}'.format(' '.join(self.keys()))

    def _ipython_display_(self):
        for identifier, tables in self.items():
            if isinstance(tables, dict) and tables:
                display_html('<h2>Tabular Data for {}</h2>'.format(identifier), raw=True)
                display_html(tables)



import uuid, json, os
from collections import OrderedDict as _OrderedDict
from collections import Mapping as _Mapping
from mpcontribs.io import replacements, mp_level01_titles

try:
    from propnet.core.quantity import Quantity
except ImportError:
    Quantity = None

class RecursiveDict(_OrderedDict):
    """extension of dict for internal representation of MPFile"""

    def rec_update(self, other=None, overwrite=False, replace_newlines=True):
        """https://gist.github.com/Xjs/114831"""
        # overwrite=False: don't overwrite existing unnested key
        if other is None: # mode to force RecursiveDicts to be used
            other = self
            overwrite = True
        for key,value in other.items():
            if isinstance(key, str):
                key = ''.join([replacements.get(c, c) for c in key])
            if key in self and \
               isinstance(self[key], dict) and \
               isinstance(value, dict):
                # ensure RecursiveDict and update key (w/o underscores)
                self[key] = RecursiveDict(self[key])
                replace_newlines = bool(key != mp_level01_titles[3])
                self[key].rec_update(other=value, overwrite=overwrite, replace_newlines=replace_newlines)
            elif (key in self and overwrite) or key not in self:
                if isinstance(value, str) and replace_newlines:
                    self[key] = value.replace('\n', ' ')
                else:
                    self[key] = value

    def iterate(self, nested_dict=None):
        """http://stackoverflow.com/questions/10756427/loop-through-all-nested-dictionary-values"""
        d = self if nested_dict is None else nested_dict
        if nested_dict is None:
            self.level = 0
        for key in list(d.keys()):
            value = d[key]
            if isinstance(value, _Mapping):
                if value.get('@class') == 'Structure':
                    from pymatgen import Structure
                    yield key, Structure.from_dict(value)
                    continue
                yield (self.level, key), None
                if value.get('@class') == 'Table':
                    from mpcontribs.io.core.components.tdata import Table
                    yield key, Table.from_dict(value)
                    continue
                #if Quantity is not None and value.get('@class') == 'Quantity':
                #    quantity = Quantity.from_dict(value)
                #    yield key, quantity
                #    continue
                if 'display' in value and 'value' in value: # 'unit' is optional
                    yield (self.level, key), value['display']
                    continue
                self.level += 1
                for inner_key, inner_value in self.iterate(nested_dict=value):
                    yield inner_key, inner_value
                self.level -= 1
            else:
                yield (self.level, key), value

    # insertion mechanism from https://gist.github.com/jaredks/6276032
    def __insertion(self, link_prev, key_value):
        key, value = key_value
        if link_prev[2] != key:
            if key in self:
                del self[key]
            link_next = link_prev[1]
            self._OrderedDict__map[key] = link_prev[1] = link_next[0] = [link_prev, link_next, key]
        dict.__setitem__(self, key, value)

    def insert_after(self, existing_key, key_value):
        self.__insertion(self._OrderedDict__map[existing_key], key_value)

    def insert_before(self, existing_key, key_value):
        self.__insertion(self._OrderedDict__map[existing_key][0], key_value)

    def insert_default_plot_options(self, pd_obj, k, update_plot_options=None):
        # make default plot (add entry in 'plots') for each
        # table, first column as x-column
        table_name = ''.join([replacements.get(c, c) for c in k])
        key = 'default_{}'.format(table_name)
        plots_dict = _OrderedDict([(
            mp_level01_titles[2], _OrderedDict([(
                key, _OrderedDict([
                    ('x', pd_obj.columns[0]), ('table', table_name)
                ])
            )])
        )])
        if update_plot_options is not None:
            plots_dict[mp_level01_titles[2]][key].update(update_plot_options)
        if mp_level01_titles[2] in self:
            self.rec_update(plots_dict)
        else:
          self[mp_level01_titles[2]] = plots_dict[mp_level01_titles[2]]

    def render(self):
        """use JsonHuman library to render a dictionary"""
        jdata, divid = json.dumps(self).replace('\\n', ' '), str(uuid.uuid4())
        html = f'<div id="{divid}" style="width:100%;"></div><script>'
        html += f'render_json({{divid: "{divid}", data: {jdata}}});</script>'
        return html

    def _ipython_display_(self):
        # TODO jupyterlab mimetype extension
        from IPython.display import display_html
        display_html(self.render(), raw=True)



import json, uuid
import pandas as pd
import plotly.io as pio
from plotly.io._utils import validate_coerce_fig_to_dict
from plotly.offline.offline import _get_jconfig
from plotly.utils import PlotlyJSONEncoder
from IPython.display import display_html
from mpcontribs.io import mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from mpcontribs.io.core.components.tdata import Table

class MyRenderer(pio.base_renderers.MimetypeRenderer):
    tid = None

    def to_mimebundle(self, fig):
        fig_dict = validate_coerce_fig_to_dict(fig, True)
        divid = str(uuid.uuid4())
        data = fig_dict.get('data', [])
        jdata = json.dumps(data, cls=PlotlyJSONEncoder, sort_keys=True)
        layout = fig_dict.get('layout', {})
        jlayout = json.dumps(layout, cls=PlotlyJSONEncoder, sort_keys=True)
        config = _get_jconfig(None)
        config.setdefault('responsive', True)
        jconfig = json.dumps(config)
        script = f'render_plot({{divid: "{divid}", layout: {jlayout}, config: {jconfig}'
        script += f', tid: "{self.tid}", data: {jdata}}})'
        html = f'<div><div id="{divid}"></div>'
        html += f'<script type="text/javascript">{script};</script></div>'
        return {'text/html': html}

my_renderer = MyRenderer()
pio.renderers['my_renderer'] = my_renderer
pio.renderers.render_on_display = True

class Plot(object):
    """class to hold and display single interactive graph/plot"""
    def __init__(self, table, config=None, tid=None):
        self.table = table
        self.config = config or {}
        self.tid = tid

    @classmethod
    def from_dict(cls, d):
        return cls(
            Table.from_dict(d), config=d.get('config'), tid=d['id']
        )

    def get_figure(self):
        from pandas import MultiIndex
        layout = dict(legend = dict(x=0.7, y=1), margin = dict(r=0, t=40))
        is_3d = isinstance(self.table.index, MultiIndex)
        if is_3d:
            import plotly.graph_objs as go
            from plotly import tools
            cols = self.table.columns
            ncols = 2 if len(cols) > 1 else 1
            nrows = len(cols)/ncols + len(cols)%ncols
            fig = tools.make_subplots(
                rows=nrows, cols=ncols, subplot_titles=cols, print_grid=False
            )
            for idx, col in enumerate(cols):
                series = self.table[col]
                z = [s.tolist() for _, s in series.groupby(level=0)]
                fig.append_trace(go.Heatmap(z=z, showscale=False), idx/ncols+1, idx%ncols+1)
            fig['layout'].update(layout)
        else:
            xaxis = self.config.get('x', self.table.columns[0])
            yaxis = self.config.get('y', None)
            yaxes = [yaxis] if yaxis is not None else \
                    [col for col in self.table.columns if col != xaxis]
            traces = []
            for axis in yaxes:
                if 'ₑᵣᵣ' not in axis:
                    tbl = self.table[[xaxis, axis]].replace('', pd.np.nan).dropna()
                    traces.append(dict(
                        x=tbl[xaxis].tolist(), y=tbl[axis].tolist(), name=axis
                    ))
            for trace in traces:
                err_axis = trace['name'] + 'ₑᵣᵣ'
                if err_axis in yaxes:
                    errors = self.table[err_axis].replace('', pd.np.nan).dropna()
                    trace['error_y'] = dict(
                        type='data', array=errors, visible=True
                    )
                    trace['mode'] = 'markers'
            layout.update(dict(
                xaxis = dict(title=xaxis),
                yaxis = dict(
                    title=self.config.get('ytitle'),
                    type=self.config.get('yaxis', {}).get('type', '-')
                ),
                showlegend=self.config.get('showlegend', True)
            ))
            fig = dict(data=traces, layout=layout)

        return fig

    def _ipython_display_(self):
        fig = self.get_figure()
        is_3d = False # TODO decide by heatmap
        axis = 'z' if is_3d else 'x'
        npts = len(fig.get('data')[0][axis])
        static_fig = (is_3d and npts > 150) or (not is_3d and npts > 700)
        renderers = ['jupyterlab']
        renderers.append('png' if static_fig else 'my_renderer')
        pio.renderers.default = '+'.join(renderers)
        pio.show(fig, tid=self.tid)

class Plots(RecursiveDict):
    """class to hold and display multiple interactive graphs/plots"""
    def __init__(self, tables, plotconfs):
        super(Plots, self).__init__(
            (plotconf['table'], Plot(
                plotconf, tables[plotconf['table']]
            )) for plotconf in plotconfs.values()
        )

    def __str__(self):
        return 'plots: {}'.format(' '.join(self.keys()))

    def _ipython_display_(self):
        for name, plot in self.items():
            if plot:
                display_html('<h3>{}</h3>'.format(name), raw=True)
                display_html(plot)

class GraphicalData(RecursiveDict):
    """class to hold and display all interactive graphs/plots of a MPFile"""
    def __init__(self, document):
        from mpcontribs.io.core.components.tdata import TabularData
        tdata = TabularData(document)
        super(GraphicalData, self).__init__(
            (identifier, Plots(
                tdata[identifier], content[mp_level01_titles[2]]
            )) for identifier, content in document.items()
            if mp_level01_titles[2] in content
        )

    def __str__(self):
        return 'mp-ids: {}'.format(' '.join(self.keys()))

    def _ipython_display_(self):
        for identifier, plots in self.items():
            if identifier != mp_level01_titles[0] and plots:
                display_html('<h2>Interactive Plots for {}</h2>'.format(identifier), raw=True)
                display_html(plots)



import os
from setuptools import setup

SETUP_PTH = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(SETUP_PTH, 'requirements.txt')) as f:
    required = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = 'mpcontribs-io',
    author = 'Patrick Huck',
    author_email = 'phuck@lbl.gov',
    description = "MPContribs I/O Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = 'https://github.com/materialsproject/MPContribs/tree/master/mpcontribs-io',
    packages = ['mpcontribs.io'],
    install_requires = required,
    license = 'MIT',
    zip_safe=False,
    include_package_data=True,
    use_scm_version={"root": "..", "relative_to": __file__},
    setup_requires=['setuptools_scm'],
)



from mpcontribs.io import mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from IPython.display import display_html

class Structures(RecursiveDict):
    """class to hold and display list of pymatgen structures for single mp-id"""
    def __init__(self, content):
        from pymatgen import Structure
        super(Structures, self).__init__(
            (key, Structure.from_dict(struc))
            for key, struc in content.get(mp_level01_titles[3], {}).items()
        )

    def _ipython_display_(self):
        for name, structure in self.items():
            if structure:
                display_html('<h4>{}</h4>'.format(name), raw=True)
                display_html('<p>{}</p>'.format(
                    structure.__repr__().replace('\n', '<br>').replace(' ', '&nbsp;')
                ), raw=True)

class StructuralData(RecursiveDict):
    """class to hold and display all pymatgen structures in MPFile"""
    def __init__(self, document):
        super(StructuralData, self).__init__(
            (identifier, Structures(content))
            for identifier, content in document.items()
        )

    def _ipython_display_(self):
        for identifier, sdata in self.items():
            if identifier != mp_level01_titles[0] and sdata:
                display_html('<h2>Structural Data for {}</h2>'.format(identifier), raw=True)
                display_html(sdata)
